#include <stdio.h>
#include <stdlib.h>
#include <omp.h>
#include "timing.h"


int main(int argc, char *argv[])
{
int N = atoi(argv[1]);
int nthreads = atoi(argv[2]);
unsigned long ticksB4, ticksAFT, ticks;
double freq = 2.5e9;


double *a, *b, *c;
a = (double *) malloc(N * sizeof(double) );
b = (double *) malloc(N * sizeof(double) );
c = (double *) malloc(N * sizeof(double) );

int i;

for(i=0; i<N; i++) {
	a[i] = (double) i;
	b[i] = (double) i;
	c[i] = (double) i;
}

get_ticks(ticksB4);

//storage for partial cumulative sums b[i]*c[i]
double partials[nthreads+1];
partials[0] = 0.;

#pragma omp parallel private(i) num_threads(nthreads)
{
int id = omp_get_thread_num();
int chunk_size = (N-1) / nthreads;
int start = id * chunk_size + 1;
int end = (id+1) * chunk_size + 1;

if (id == nthreads-1)
	end = N;

//each thread starts its partial sum
a[start] = b[start]*c[start];

//and iteratively adds to it
//if thread 1 is responsible for iterations 3:5, then it computes:
//a[3] = b[3]*c[3], a[4] = b[3]*c[3] + b[4]*c[4],
//a[5] = b[3]*c[3] + b[4]*c[4] + b[5]*c[5]
for(i=start+1; i<end; i++){
	a[i] = a[i-1] + b[i]*c[i];
}

//threads put their final partial sum into partials[]
partials[id+1] = a[end-1];

//wait for threads to finish
#pragma omp barrier

//only one thread accumulates the partial sums and adds a[0]
#pragma omp single
{
	for(i=2;i<nthreads+1;i++){
		partials[i] += partials[i-1];
	}
	for(i=0;i<nthreads+1;i++){
		partials[i] += a[0];
	}
}

//then each thread goes back over its region of a[] and adds
//the partial sums from all threads before it
for(i=start; i<end; i++){
	a[i] = a[i] + partials[id];
}

}


get_ticks(ticksAFT);
ticks = ticksAFT - ticksB4;


printf("\na:");
for(i=0; i<N; i++) {
	printf(" %.3g", a[i]);
}
printf("\n\n");


printf("runtime (s): %g\n\n", ticks/freq);

return 0;
}
